#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
# Edits by Kyle Bystrom <kylebystrom@gmail.com>
#

import time

import numpy
from pyscf.sgx.sgx_jk import _gen_batch_nuc, _gen_jk_direct, lib, logger


def get_jk_densities(sgx, dm, hermi=1, direct_scf_tol=1e-13):
    t0 = time.monotonic(), time.time()
    mol = sgx.mol
    grids = sgx.grids
    gthrd = sgx.grids_thrd

    dms = numpy.asarray(dm)
    dm_shape = dms.shape
    nao = dm_shape[-1]
    dms = dms.reshape(-1, nao, nao)
    nset = dms.shape[0]

    if sgx.debug:
        batch_nuc = _gen_batch_nuc(mol)
    else:
        batch_jk = _gen_jk_direct(
            mol, "s2", True, True, direct_scf_tol, sgx._opt, pjs=False
        )
    logger.timer_debug1(mol, "sgX initialziation", *t0)

    ngrids = grids.coords.shape[0]
    max_memory = sgx.max_memory - lib.current_memory()[0]
    sblk = sgx.blockdim
    blksize = min(ngrids, max(4, int(min(sblk, max_memory * 1e6 / 8 / nao**2))))
    tnuc = 0, 0
    ej, ek = numpy.zeros((nset, ngrids)), numpy.zeros((nset, ngrids))
    for i0, i1 in lib.prange(0, ngrids, blksize):
        coords = grids.coords[i0:i1]
        ao = mol.eval_gto("GTOval", coords)
        wao = ao * grids.weights[i0:i1, None]

        fg = lib.einsum("gi,xij->xgj", wao, dms)
        fgnw = lib.einsum("gi,xij->xgj", ao, dms)
        mask = numpy.zeros(i1 - i0, dtype=bool)
        for i in range(nset):
            mask |= numpy.any(fg[i] > gthrd, axis=1)
            mask |= numpy.any(fg[i] < -gthrd, axis=1)

        inds = numpy.arange(i0, i1)

        numpy.einsum("xgu,gu->xg", fg, ao)
        rho = numpy.einsum("xgu,gu->xg", fgnw, ao)

        if sgx.debug:
            tnuc = tnuc[0] - time.monotonic(), tnuc[1] - time.time()
            gbn = batch_nuc(mol, coords)
            tnuc = tnuc[0] + time.monotonic(), tnuc[1] + time.time()
            jg = numpy.einsum("gij,xij->xg", gbn, dms)
            gv = lib.einsum("gvt,xgt->xgv", gbn, fgnw)
            gbn = None
        else:
            tnuc = tnuc[0] - time.monotonic(), tnuc[1] - time.time()
            jg, gv = batch_jk(mol, coords, dms, fgnw.copy(), None)
            tnuc = tnuc[0] + time.monotonic(), tnuc[1] + time.time()

        jg = numpy.sum(jg, axis=0)
        ej[:, inds] = jg * rho / 2.0
        for i in range(nset):
            ek[i, inds] = lib.einsum("gu,gu->g", fgnw[i], gv[i])

        jg = gv = None

    ek *= -0.5
    return ej, ek



import numpy as np
from gpaw.xc.kernel import XCKernel
from gpaw.xc.libxc import LibXC

from ciderpress.dft.plans import SemilocalPlan2
from ciderpress.gpaw.config import GPAW_DEFAULT_RHO_TOL


class CiderKernel(XCKernel):
    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    def todict(self):
        raise NotImplementedError

    def call_xc_kernel(
        self,
        e_g,
        n_sg,
        sigma_xg,
        feat_sg,
        v_sg,
        dedsigma_xg,
        vfeat_sg,
        tau_sg=None,
        dedtau_sg=None,
    ):
        # make view so we can reshape things
        e_g = e_g.view()
        n_sg = n_sg.view()
        v_sg = v_sg.view()
        feat_sg = feat_sg.view()
        vfeat_sg = vfeat_sg.view()
        if tau_sg is not None:
            tau_sg = tau_sg.view()
            dedtau_sg = dedtau_sg.view()

        nspin = n_sg.shape[0]
        ngrids = n_sg.size // nspin
        nfeat = self.mlfunc.settings.nfeat
        X0T = np.empty((nspin, nfeat, ngrids))
        has_sl = True
        has_nldf = not self.mlfunc.settings.nldf_settings.is_empty
        start = 0
        sigma_sg = sigma_xg[::2]
        dedsigma_sg = dedsigma_xg[::2].copy()
        new_shape = (nspin, ngrids)
        e_g.shape = (ngrids,)
        n_sg.shape = new_shape
        v_sg.shape = new_shape
        sigma_sg.shape = new_shape
        dedsigma_sg.shape = new_shape
        feat_sg.shape = (nspin, -1, ngrids)
        vfeat_sg.shape = (nspin, -1, ngrids)
        if has_sl:
            sl_plan = SemilocalPlan2(self.mlfunc.settings.sl_settings, nspin)
        else:
            sl_plan = None
        if tau_sg is not None:
            tau_sg.shape = new_shape
            dedtau_sg.shape = new_shape
        if has_sl:
            nfeat_tmp = self.mlfunc.settings.sl_settings.nfeat
            X0T[:, start : start + nfeat_tmp] = sl_plan.get_feat(
                n_sg, sigma_sg, tau=tau_sg
            )
            start += nfeat_tmp
        if has_nldf:
            nfeat_tmp = self.mlfunc.settings.nldf_settings.nfeat
            # X0T[:, start : start + nfeat_tmp] = nspin * feat_sg
            X0T[:, start : start + nfeat_tmp] = feat_sg
            start += nfeat_tmp
        X0TN = self.mlfunc.settings.normalizers.get_normalized_feature_vector(X0T)
        exc_ml, dexcdX0TN_ml = self.mlfunc(X0TN, rhocut=self.rhocut)
        xmix = self.xmix  # / rho.shape[0]
        exc_ml *= xmix
        dexcdX0TN_ml *= xmix
        vxc_ml = self.mlfunc.settings.normalizers.get_derivative_wrt_unnormed_features(
            X0T, dexcdX0TN_ml
        )
        # vxc_ml = dexcdX0TN_ml
        e_g[:] += exc_ml

        start = 0
        if has_sl:
            nfeat_tmp = self.mlfunc.settings.sl_settings.nfeat
            sl_plan.get_vxc(
                vxc_ml[:, start : start + nfeat_tmp],
                n_sg,
                v_sg,
                sigma_sg,
                dedsigma_sg,
                tau=tau_sg,
                vtau=dedtau_sg,
            )
            start += nfeat_tmp
        if has_nldf:
            nfeat_tmp = self.mlfunc.settings.nldf_settings.nfeat
            # vfeat_sg[:] += nspin * vxc_ml[:, start : start + nfeat_tmp]
            vfeat_sg[:] += vxc_ml[:, start : start + nfeat_tmp]
            start += nfeat_tmp
        dedsigma_xg[::2] = dedsigma_sg.reshape(nspin, *dedsigma_xg.shape[1:])

    def calculate(self, *args):
        raise NotImplementedError


class CiderGGAHybridKernel(CiderKernel):
    def __init__(self, mlfunc, xmix, xstr, cstr, rhocut=GPAW_DEFAULT_RHO_TOL):
        self.type = "GGA"
        self.name = "CiderGGA"
        self.xkernel = LibXC(xstr)
        self.ckernel = LibXC(cstr)
        self.xmix = xmix
        self.mlfunc = mlfunc
        self.rhocut = rhocut

    def todict(self):
        return {
            "xmix": self.xmix,
            "xstr": self.xkernel.name,
            "cstr": self.ckernel.name,
        }

    def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg, feat_sg):
        """
        Evaluate CIDER 'hybrid' functional.
        """
        n_sg.shape[0]
        self.xkernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
        e_g[:] *= 1 - self.xmix
        v_sg[:] *= 1 - self.xmix
        dedsigma_xg[:] *= 1 - self.xmix
        e_g_tmp, v_sg_tmp, dedsigma_xg_tmp = (
            np.zeros_like(e_g),
            np.zeros_like(v_sg),
            np.zeros_like(dedsigma_xg),
        )
        self.ckernel.calculate(e_g_tmp, n_sg, v_sg_tmp, sigma_xg, dedsigma_xg_tmp)
        e_g[:] += e_g_tmp
        v_sg[:] += v_sg_tmp
        dedsigma_xg[:] += dedsigma_xg_tmp
        vfeat_sg = np.zeros_like(feat_sg)
        self.call_xc_kernel(
            e_g,
            n_sg,
            sigma_xg,
            feat_sg,
            v_sg,
            dedsigma_xg,
            vfeat_sg,
        )
        return vfeat_sg


class CiderMGGAHybridKernel(CiderGGAHybridKernel):
    def __init__(self, mlfunc, xmix, xstr, cstr, rhocut=GPAW_DEFAULT_RHO_TOL):
        self.type = "MGGA"
        self.name = "CiderMGGA"
        self.xkernel = LibXC(xstr)
        self.ckernel = LibXC(cstr)
        self.xmix = xmix
        self.mlfunc = mlfunc
        self.rhocut = rhocut

    def calculate(
        self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg, tau_sg, dedtau_sg, feat_sg
    ):
        """
        Evaluate CIDER 'hybrid' functional.
        """
        n_sg.shape[0]
        if self.xkernel.type == "GGA":
            self.xkernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
            dedtau_sg[:] = 0
        else:
            self.xkernel.calculate(
                e_g, n_sg, v_sg, sigma_xg, dedsigma_xg, tau_sg, dedtau_sg
            )
        e_g[:] *= 1 - self.xmix
        v_sg[:] *= 1 - self.xmix
        dedsigma_xg[:] *= 1 - self.xmix
        dedtau_sg[:] *= 1 - self.xmix
        e_g_tmp, v_sg_tmp, dedsigma_xg_tmp, dedtau_sg_tmp = (
            np.zeros_like(e_g),
            np.zeros_like(v_sg),
            np.zeros_like(dedsigma_xg),
            np.zeros_like(dedtau_sg),
        )
        if self.ckernel.type == "GGA":
            self.ckernel.calculate(e_g_tmp, n_sg, v_sg_tmp, sigma_xg, dedsigma_xg_tmp)
        else:
            self.ckernel.calculate(
                e_g_tmp,
                n_sg,
                v_sg_tmp,
                sigma_xg,
                dedsigma_xg_tmp,
                tau_sg,
                dedtau_sg_tmp,
            )
        e_g[:] += e_g_tmp
        v_sg[:] += v_sg_tmp
        dedsigma_xg[:] += dedsigma_xg_tmp
        dedtau_sg[:] += dedtau_sg_tmp
        vfeat_sg = np.zeros_like(feat_sg)
        self.call_xc_kernel(
            e_g,
            n_sg,
            sigma_xg,
            feat_sg,
            v_sg,
            dedsigma_xg,
            vfeat_sg,
            tau_sg=tau_sg,
            dedtau_sg=dedtau_sg,
        )
        return vfeat_sg



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np

from ciderpress.dft.baselines import get_libxc_baseline
from ciderpress.dft.xc_evaluator import FuncEvaluator, XCEvalSerializable


class KernelEvalBase2:

    mode = None
    feature_list = None
    _mul_basefunc: str = None
    _add_basefunc: str = None

    @property
    def N1(self):
        raise NotImplementedError

    def get_descriptors(self, X0T, force_polarize=False):
        """
        Compute and return transformed descriptor matrix X1.

        Args:
            rho: Density/Kinetic energy density representation.
            X0 (nspin, N0, Nsamp): Raw features
            nspin (1 or 2): Number of spins

        Returns:
            X1 (Nsamp, N1)
        """
        nspin, N0, Nsamp = X0T.shape
        if force_polarize and self.mode == "POL" and nspin == 1:
            X0T = np.concatenate([X0T, X0T], axis=0)
            nspin = 2
        N1 = self.N1
        if self.mode == "SEP" or self.mode == "POL":
            X1 = np.zeros((nspin, Nsamp, N1))
            for s in range(nspin):
                self.feature_list.fill_vals_(X1[s].T, X0T[s])
            X1 = X1.reshape(nspin * Nsamp, N1)
        elif self.mode == "NPOL":
            X0T_sum = X0T.mean(0)
            X1 = np.zeros((Nsamp, N1))
            self.feature_list.fill_vals_(X1.T, X0T_sum)
        else:
            raise NotImplementedError
        if force_polarize and self.mode == "POL":
            return X1.reshape(nspin, Nsamp, N1)
        return X1

    def get_descriptors_with_mul(self, X0T, multiplier):
        """
        Compute and return transformed descriptor matrix X1.

        Args:
            rho: Density/Kinetic energy density representation.
            X0 (nspin, N0, Nsamp): Raw features
            nspin (1 or 2): Number of spins

        Returns:
            X1 (Nsamp, N1)
        """
        nspin, N0, Nsamp = X0T.shape
        N1 = self.N1
        if self.mode == "SEP" or self.mode == "POL":
            X1 = np.zeros((nspin, Nsamp, N1))
            for s in range(nspin):
                self.feature_list.fill_vals_(X1[s].T, X0T[s])
                X1[s] *= multiplier[:, None]
            X1 = X1.reshape(nspin * Nsamp, N1)
        elif self.mode == "NPOL":
            X0T_sum = X0T.mean(0)
            X1 = np.zeros((Nsamp, N1))
            self.feature_list.fill_vals_(X1.T, X0T_sum)
            X1[:] *= multiplier[:, None]
        else:
            raise NotImplementedError
        return X1

    def _get_baseline(self, xcid, rho_tuple):
        if self.mode == "SEP":
            nspin = rho_tuple[0].shape[0]
            if nspin == 1:
                res = list(get_libxc_baseline(xcid, rho_tuple))
                res[0] = res[0][None, :]
                return tuple(res)
            else:
                assert nspin == 2
                sep_res = [np.zeros_like(rho_tuple[0]), np.zeros_like(rho_tuple[0])]
                if len(rho_tuple) > 1:
                    sep_res.append(np.zeros_like(rho_tuple[1]))
                if len(rho_tuple) > 2:
                    sep_res.append(np.zeros_like(rho_tuple[2]))
                sep_res = tuple(sep_res)
                for s in range(nspin):
                    tuple_s = [2 * rho_tuple[0][s : s + 1]]
                    if len(rho_tuple) > 1:
                        tuple_s.append(4 * rho_tuple[1][2 * s : 2 * s + 1])
                    if len(rho_tuple) > 2:
                        tuple_s.append(2 * rho_tuple[2][s : s + 1])
                    res = get_libxc_baseline(xcid, tuple_s)
                    sep_res[0][s] = 0.5 * res[0]
                    sep_res[1][s] = res[1]
                    if len(res) > 2:
                        sep_res[2][2 * s] = 2 * res[2]
                    if len(res) > 3:
                        sep_res[3][s] = res[3]
                return sep_res
        else:
            return get_libxc_baseline(xcid, rho_tuple)

    def multiplicative_baseline(self, rho_tuple):
        return self._get_baseline(self._mul_basefunc, rho_tuple)

    def additive_baseline(self, rho_tuple):
        if self._add_basefunc is None:
            return None
        return self._get_baseline(self._add_basefunc, rho_tuple)

    def apply_descriptor_grad(self, X0T, dfdX1, force_polarize=False):
        """

        Args:
            X0T (nspin, N0, Nsamp): raw feature descriptors
            dfdX1 (nspin * Nsamp, N1): derivative with respect to transformed
                descriptors X1

        Returns:
            dfdX0T (nspin, N0, Nsamp): derivative with respect
                to raw descriptors X0.
        """
        nspin, N0, Nsamp = X0T.shape
        N1 = self.N1
        if force_polarize and dfdX1.shape[0] == 2 and nspin == 1:
            dfdX1 = dfdX1[:1]
        if self.mode == "SEP" or self.mode == "POL":
            dfdX0T = np.zeros_like(X0T)
            dfdX1 = dfdX1.reshape(nspin, Nsamp, N1)
            for s in range(nspin):
                self.feature_list.fill_derivs_(dfdX0T[s], dfdX1[s].T, X0T[s])
        elif self.mode == "NPOL":
            dfdX0T = np.zeros_like(X0T)
            self.feature_list.fill_derivs_(dfdX0T[0], dfdX1.T, X0T.mean(0))
            for s in range(1, nspin):
                dfdX0T[s] = dfdX0T[0]
            dfdX0T /= nspin
        else:
            raise NotImplementedError
        return dfdX0T

    def apply_libxc_baseline_(self, f, dfdX1, rho_tuple, vrho_tuple, add_base=True):
        m_res = self.multiplicative_baseline(rho_tuple)
        add_base = add_base and self._add_basefunc is not None
        for vrho, vm in zip(vrho_tuple, m_res[1:]):
            if self.mode == "SEP" and vrho.shape[0] == 3:
                vrho[::2] += vm[::2] * f
            else:
                vrho[:] += vm * f
        if self.mode == "SEP":
            assert m_res[0].shape == f.shape == dfdX1.shape[:2]
        elif self.mode == "POL":
            assert m_res[0].shape == f.shape == (dfdX1.shape[1],)
        else:
            assert m_res[0].shape == f.shape == (dfdX1.shape[0],)
        dfdX1[:] *= m_res[0][..., None]
        f[:] *= m_res[0]
        if add_base:
            a_res = self.additive_baseline(rho_tuple)
            f[:] += a_res[0]
            for vrho, va in zip(vrho_tuple, a_res[1:]):
                vrho[:] += va


class MappedDFTKernel2(KernelEvalBase2, XCEvalSerializable):
    """
    This class evaluates the XC term arising from a single DFTKernel
    object, using one or a list of FuncEvaluator objects.
    """

    def __init__(
        self,
        fevals,
        feature_list,
        mode,
        multiplicative_baseline,
        additive_baseline=None,
    ):
        """
        fevals: FuncEvaluator
        """
        self.fevals = fevals
        if not isinstance(self.fevals, list):
            self.fevals = [self.fevals]
        for feval in self.fevals:
            if not isinstance(feval, FuncEvaluator):
                raise ValueError
        self.mode = mode
        self.feature_list = feature_list
        self._mul_basefunc = multiplicative_baseline
        self._add_basefunc = additive_baseline

    @property
    def N1(self):
        return self.feature_list.nfeat

    def __call__(self, X0T, rho_tuple, vrho_tuple, rhocut=0):
        X1 = self.get_descriptors(X0T, force_polarize=True)
        Nsamp_internal = X1.shape[-2]
        f = np.zeros(Nsamp_internal)
        df = np.zeros_like(X1)
        for feval in self.fevals:
            feval(X1, f, df)
        if self.mode == "POL" and X0T.shape[0] == 1:
            df = 2 * df[:1]
        if self.mode == "SEP":
            f = f.reshape(X0T.shape[0], -1)
            df = df.reshape(X0T.shape[0], -1, self.N1)
        if rhocut > 0:
            cond = rho_tuple[0] < rhocut
            if self.mode == "SEP":
                f[cond] = 0.0
                df[cond] = 0.0
            else:
                scond = rho_tuple[0].sum(0) < rhocut
                f[scond] = 0.0
                if self.mode == "POL":
                    df[cond, :] = 0.0
                else:
                    df[scond, :] = 0.0
        self.apply_libxc_baseline_(f, df, rho_tuple, vrho_tuple)
        dfdX0T = self.apply_descriptor_grad(X0T, df, force_polarize=True)
        if self.mode == "SEP":
            f = f.sum(0)
        return f, dfdX0T

    def to_dict(self):
        return {
            "fevals": [fe.to_dict() for fe in self.fevals],
            "feature_list": self.feature_list.to_dict(),
            "mode": self.mode,
        }

    @classmethod
    def from_dict(cls, d):
        raise NotImplementedError


class MappedXC2:
    def __init__(self, mapped_kernels, settings, libxc_baseline=None):
        """

        Args:
            mapped_kernels (list of MappedDFTKernel2):
            settings (FeatureSettings):
            libxc_baseline (str):
        """
        self.kernels = mapped_kernels
        self.settings = settings
        self.libxc_baseline = libxc_baseline

    def set_baseline_mode(self, mode):
        # TODO baseline can be GPAW or PySCF mode.
        # Need to implement for more complicated XC.
        raise NotImplementedError

    def __call__(self, X0T, rho_tuple, vrho_tuple=None, rhocut=0):
        """
        Evaluate functional from normalized features
        Args:
            X0T (array): normalized features
            rhocut (float): low-density cutoff

        Returns:
            res, dres (array, array), XC energy contribution
                and functional derivative
        """
        vrho_tuple = tuple([np.zeros_like(r, order="F") for r in rho_tuple])
        res, dres = 0, 0
        for kernel in self.kernels:
            tmp, dtmp = kernel(X0T, rho_tuple, vrho_tuple, rhocut=rhocut)
            res += tmp
            dres += dtmp
        return res, dres, vrho_tuple

    @property
    def normalizer(self):
        return self.settings.normalizers

    @property
    def nfeat(self):
        return self.normalizer.nfeat



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

# sklearn kernels for CIDER models.


from inspect import signature

import numpy as np
from scipy.special import comb
from sklearn.gaussian_process.kernels import (
    RBF,
    ConstantKernel,
    DotProduct,
    Exponentiation,
    GenericKernelMixin,
    Hyperparameter,
    Kernel,
    Product,
    StationaryKernelMixin,
    Sum,
    WhiteKernel,
    _num_samples,
    cdist,
)


def _check_length_scale(X, length_scale):
    length_scale = np.squeeze(length_scale).astype(float)
    if np.ndim(length_scale) > 1:
        raise ValueError("length_scale cannot be of dimension greater than 1")
    if np.ndim(length_scale) == 1 and X.shape[1] != length_scale.shape[0]:
        raise ValueError(
            "Anisotropic kernel must have the same number of "
            "dimensions as data (%d!=%d)" % (length_scale.shape[0], X.shape[1])
        )
    return length_scale


class DiffKernelMixin:
    def __add__(self, b):
        if not isinstance(b, Kernel):
            return DiffSum(self, DiffConstantKernel(b))
        return DiffSum(self, b)

    def __radd__(self, b):
        if not isinstance(b, Kernel):
            return DiffSum(DiffConstantKernel(b), self)
        return DiffSum(b, self)

    def __mul__(self, b):
        if not isinstance(b, Kernel):
            return DiffProduct(self, DiffConstantKernel(b))
        return DiffProduct(self, b)

    def __rmul__(self, b):
        if not isinstance(b, Kernel):
            return DiffProduct(DiffConstantKernel(b), self)
        return DiffProduct(b, self)

    def __pow__(self, b):
        return DiffExponentiation(self, b)

    def k_and_deriv(self, X, Y=None):
        raise NotImplementedError


class DiffSum(DiffKernelMixin, Sum):
    def k_and_deriv(self, X, Y=None):
        k1, dk1 = self.k1.k_and_deriv(X, Y)
        k2, dk2 = self.k2.k_and_deriv(X, Y)
        return k1 + k2, dk1 + dk2


class DiffProduct(DiffKernelMixin, Product):
    def k_and_deriv(self, X, Y=None):
        k1, dk1 = self.k1.k_and_deriv(X, Y)
        k2, dk2 = self.k2.k_and_deriv(X, Y)
        return k1 * k2, k1[..., None] * dk2 + k2[..., None] * dk1


class DiffExponentiation(DiffKernelMixin, Exponentiation):
    def k_and_deriv(self, X, Y=None):
        k, dk = self.kernel.k_and_deriv(X, Y)
        return (
            k**self.exponent,
            (self.exponent * k ** (self.exponent - 1))[:, :, np.newaxis] * dk,
        )


class DiffTransform(DiffKernelMixin, Kernel):
    def __init__(self, kernel, matrix, std=None, avg=None):
        self.kernel = kernel
        self.matrix = matrix
        self.std = std
        self.avg = avg

    def get_params(self, deep=True):
        params = dict(kernel=self.kernel, exponent=self.exponent)
        if deep:
            deep_items = self.kernel.get_params().items()
            params.update(("kernel__" + k, val) for k, val in deep_items)
        return params

    @property
    def hyperparameters(self):
        """Returns a list of all hyperparameter."""
        r = []
        for hyperparameter in self.kernel.hyperparameters:
            r.append(
                Hyperparameter(
                    "kernel__" + hyperparameter.name,
                    hyperparameter.value_type,
                    hyperparameter.bounds,
                    hyperparameter.n_elements,
                )
            )
        return r

    @property
    def theta(self):
        return self.kernel.theta

    @theta.setter
    def theta(self, theta):
        self.kernel.theta = theta

    @property
    def bounds(self):
        return self.kernel.bounds

    def __eq__(self, b):
        if type(self) is not type(b):
            return False
        return self.kernel == b.kernel and self.exponent == b.exponent

    def _transform(self, X):
        if X is None:
            return None
        if self.avg is not None:
            X = X - self.avg
        if self.std is not None:
            X = X / self.std
        return X.dot(self.matrix)

    def _transform_bwd(self, X):
        if X is None:
            return None
        X = X.dot(self.matrix.T)
        if self.std is not None:
            X = X / self.std
        return X

    def __call__(self, X, Y=None, eval_gradient=False):
        return self.kernel(
            self._transform(X), Y=self._transform(Y), eval_gradient=eval_gradient
        )

    def diag(self, X):
        return self.kernel.diag(self._transform(X))

    def k_and_deriv(self, X, Y=None):
        k, dk = self.kernel.k_and_deriv(self._transform(X), self._transform(Y))
        return k, self._transform_bwd(dk)

    def __repr__(self):
        return "{0} ** {1}".format(self.kernel, self.exponent)

    def is_stationary(self):
        """Returns whether the kernel is stationary."""
        return self.kernel.is_stationary()

    @property
    def requires_vector_input(self):
        """Returns whether the kernel is defined on discrete structures."""
        return self.kernel.requires_vector_input


class DiffWhiteKernel(DiffKernelMixin, WhiteKernel):
    def k_and_deriv(self, X, Y=None):
        k = self.__call__(X, Y)
        shape = k.shape + (X.shape[1],)
        return k, np.zeros(shape, dtype=k.dtype)


class DiffConstantKernel(DiffKernelMixin, ConstantKernel):
    def k_and_deriv(self, X, Y=None):
        k = self.__call__(X, Y)
        shape = k.shape + (X.shape[1],)
        return k, np.zeros(shape, dtype=k.dtype)


class DiffRBF(DiffKernelMixin, RBF):
    def k_and_deriv(self, X, Y=None):
        """
        Compute the kernel and its derivative with
        respect to the input X.

        Args:
            X: Inputs with respect to which the
               kernel derivative is evaluated.
            Y: Other input. If None, set to X. Note:
                If Y is None i.e. X, Y is still assumed
                stationary, i.e. the return value is
                0.5 * d k(X,X,) / dX

        Returns:
            k, (X.shape[0], Y.shape[0])
            dk (X.shape[0], Y.shape[0], X.shape[1])
        """
        if Y is None:
            Y = X
        k = self.__call__(X, Y)
        dk = k[:, :, None] * (Y[None, :, :] - X[:, None, :])
        dk /= self.length_scale**2
        return k, dk


class DiffAntisymRBF(DiffRBF):
    """
    This is like a regular RBF kernel, except it obeys a sort
    of "antisymmetry" property where
    k(x0, x1, x2, ....; x0', x1', x2' ....) = 0
    if x0 == x1 or x0' == x1'.
    This property applies only to the first two features. It could
    be useful for enforcing exact constraints.
    """

    def __call__(self, X, Y=None, eval_gradient=False):
        if eval_gradient:
            raise NotImplementedError(
                "eval_gradient not implemented for this kernel yet"
            )
        length_scale = _check_length_scale(X[:, 1:], self.length_scale)
        if Y is None:
            Y = X.copy()
        XT = X[:, 2:] / length_scale[1:]
        YT = Y[:, 2:] / length_scale[1:]
        dists = cdist(XT, YT, metric="sqeuclidean")
        KT = np.exp(-0.5 * dists)
        XS = X[:, :2] / length_scale[0]
        YS = Y[:, :2] / length_scale[0]
        dists = cdist(XS[:, 0:1], YS[:, 0:1], metric="sqeuclidean")
        KS = np.exp(-0.5 * dists)
        dists = cdist(XS[:, 0:1], YS[:, 1:2], metric="sqeuclidean")
        KS[:] -= np.exp(-0.5 * dists)
        dists = cdist(XS[:, 1:2], YS[:, 0:1], metric="sqeuclidean")
        KS[:] -= np.exp(-0.5 * dists)
        dists = cdist(XS[:, 1:2], YS[:, 1:2], metric="sqeuclidean")
        KS[:] += np.exp(-0.5 * dists)
        return KS * KT

    def k_and_deriv(self, X, Y=None):
        length_scale = _check_length_scale(X[:, 1:], self.length_scale)
        if Y is None:
            Y = X.copy()
        XT = X[:, 2:] / length_scale[1:]
        YT = Y[:, 2:] / length_scale[1:]
        DK = np.zeros((XT.shape[0], YT.shape[0], X.shape[1]))
        dists = cdist(XT, YT, metric="sqeuclidean")
        KT = np.exp(-0.5 * dists)
        XS = X[:, :2] / length_scale[0]
        YS = Y[:, :2] / length_scale[0]

        dists = cdist(XS[:, 0], YS[:, 0], metric="sqeuclidian")
        tmp = np.exp(-0.5 * dists)
        DK[..., 0] = tmp * (YS[:, 0] - XS[:, 0])
        KS = tmp

        dists = cdist(XS[:, 0], YS[:, 1], metric="sqeuclidian")
        tmp = np.exp(-0.5 * dists)
        DK[..., 0] += tmp * (XS[:, 0] - YS[:, 1])
        KS[:] -= tmp

        dists = cdist(XS[:, 1], YS[:, 0], metric="sqeuclidian")
        tmp = np.exp(-0.5 * dists)
        DK[..., 1] = tmp * (XS[:, 1] - YS[:, 0])
        KS[:] -= tmp

        dists = cdist(XS[:, 1], YS[:, 1], metric="sqeuclidian")
        tmp = np.exp(-0.5 * dists)
        DK[..., 1] += tmp * (YS[:, 1] - XS[:, 1])
        KS[:] += tmp

        DK[..., 2:] = (KS * KT)[..., None] * (YT[None, :, :] - XT[:, None, :])
        DK[..., 2:] /= length_scale[1:]
        DK[..., :2] /= length_scale[0]
        return KS * KT, DK


class DiffLinearKernel(DiffKernelMixin, Kernel):
    def __call__(self, X, Y=None, eval_gradient=False):
        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError
        k = X.dot(Y.T)
        if eval_gradient:
            dk = np.empty((X.shape[0], X.shape[0], 0))
            return k, dk
        else:
            return k

    def diag(self, X):
        return np.einsum("if,if->i", X, X)

    def is_stationary(self):
        return False

    def k_and_deriv(self, X, Y=None):
        if Y is None:
            Y = X
        k = self.__call__(X, Y)
        return k, np.tile(Y, (X.shape[0], 1, 1))


class DiffPolyKernel(DiffKernelMixin, Kernel):
    def __init__(self, gamma=1.0, gamma_bounds=(1e-5, 1e5), order=4, factorial=True):
        self.gamma = gamma
        self.gamma_bounds = gamma_bounds
        self.order = order
        self.factorial = factorial

    @property
    def anisotropic(self):
        return np.iterable(self.gamma) and len(self.gamma) > 1

    @property
    def hyperparameter_gamma(self):
        if self.anisotropic:
            return Hyperparameter(
                "gamma", "numeric", self.gamma_bounds, len(self.gamma)
            )
        return Hyperparameter("gamma", "numeric", self.gamma_bounds)

    def __call__(self, X, Y=None, eval_gradient=False):
        k = 1.0
        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError
        optg = not self.hyperparameter_gamma.fixed
        if eval_gradient and optg:
            dk = 0
        dot1 = (self.gamma * X).dot(Y.T)
        dotn = 1
        for n in range(1, self.order + 1):
            if self.factorial:
                if eval_gradient and optg:
                    dk += dotn
                dotn *= dot1 / n
            else:
                if eval_gradient and optg:
                    dk += n * dotn
                dotn *= dot1
            k += dotn
        if eval_gradient:
            if optg:
                if self.anisotropic:
                    dk = dk[:, :, None] * X[:, None, :] * X[None, :, :] * self.gamma
                else:
                    dk = (dk * dot1)[:, :, np.newaxis]
            else:
                dk = np.empty((X.shape[0], X.shape[0], 0), dtype=X.dtype)
            return k, dk
        return k

    def diag(self, X):
        k = 1.0
        dot1 = np.einsum("ij,ij->i", self.gamma * X, X)
        dotn = 1
        for n in range(1, self.order + 1):
            if self.factorial:
                dotn *= dot1 / n
            else:
                dotn *= dot1
            k += dotn
        return k

    def is_stationary(self):
        return False

    def k_and_deriv(self, X, Y=None):
        if Y is None:
            Y = X
        k = 1.0
        dk = 0.0
        dot1 = (self.gamma * X).dot(Y.T)
        dotn = 1
        for n in range(1, self.order + 1):
            if self.factorial:
                dk += dotn
                dotn *= dot1 / n
            else:
                dk += dotn * n
                dotn *= dot1
            k += dotn
        dk = self.gamma * dk[:, :, None] * Y[None, :, :]
        return k, dk


class _IndexMixin:
    def get_params(self, deep=True):
        params = dict()
        cls = self._base_cls
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        init_sign = signature(init)
        args, varargs = [], []
        for parameter in init_sign.parameters.values():
            if parameter.kind != parameter.VAR_KEYWORD and parameter.name != "self":
                args.append(parameter.name)
            if parameter.kind == parameter.VAR_POSITIONAL:
                varargs.append(parameter.name)

        if len(varargs) != 0:
            raise RuntimeError(
                "scikit-learn kernels should always "
                "specify their parameters in the signature"
                " of their __init__ (no varargs)."
                " %s doesn't follow this convention." % (cls,)
            )
        for arg in args:
            params[arg] = getattr(self, arg)
        if deep:
            param_keys = list(params.keys())
            for k in param_keys:
                if hasattr(params[k], "get_params"):
                    deep_items = params[k].get_params().items()
                    params.update(
                        (("{}__{}".format(k, kk), val) for kk, val in deep_items)
                    )

        return params

    @property
    def hyperparameters(self):
        """Returns a list of all hyperparameter specifications."""
        r = [
            getattr(self, attr)
            for attr in dir(self)
            if attr.startswith("hyperparameter_")
        ]
        for attr in self.get_params():
            if hasattr(attr, "hyperparameters"):
                for hparam in attr.hyperparameters:
                    r.append(hparam)
        return r


class _SubsetMixin(_IndexMixin):
    """
    A Mixin that evalutes the kernel on a subset of features.
    To use, make a class SubsetClass(_SubsetMixin, BaseClass).
    BaseClass must be the second parent.
    """

    def __init__(self, indexes, *args, **kwargs):
        """
        Initialize SubsetKernel
        Args:
            indexes: list/tuple of indexes or slice
            *args and **kwargs: arguments to parent kernel
        """
        self._locked = False
        self.indexes = indexes
        self._base_cls = self.__class__.__bases__[1]  # Base Kernel
        self._base_cls.__init__(self, *args, **kwargs)

    def _index_and_lock(self, X):
        if self._locked:
            return False, X
        else:
            self._locked = True
            return True, X[:, self.indexes]

    def __call__(self, X, Y=None, eval_gradient=False, **kwargs):
        """
        Call the kernel

        Args:
            X: First matrix
            Y: Second matrix
            eval_gradient: Whether to evaluate gradient. X must be None
            **kwargs: Passed to base kernel. Primarily for using the
                get_sub_kernels option of ARBF for spline mapping.

        Returns:
            Return values of base kernel (usually kernel matrix and, if
                eval_gradient is True, its gradient wrt hyperparams)
        """
        if self._locked:
            return self._base_cls.__call__(
                self, X, Y=Y, eval_gradient=eval_gradient, **kwargs
            )
        self._locked = True
        if Y is not None:
            Y = Y[:, self.indexes]
        X = X[:, self.indexes]
        result = self._base_cls.__call__(
            self, X, Y=Y, eval_gradient=eval_gradient, **kwargs
        )
        self._locked = False
        return result

    def diag(self, X):
        if self._locked:
            return self._base_cls.diag(self, X)
        self._locked = True
        result = self._base_cls.diag(self, X[:, self.indexes])
        self._locked = False
        return result

    def k_and_deriv(self, X, Y=None):
        if self._locked:
            return self._base_cls.k_and_deriv(self, X, Y=Y)
        self._locked = True
        if Y is not None:
            Y = Y[:, self.indexes]
            shape = (X.shape[0], Y.shape[0], X.shape[1])
        else:
            shape = (X.shape[0], X.shape[0], X.shape[1])
        dk = np.zeros(shape)
        X = X[:, self.indexes]
        k, dk[:, :, self.indexes] = self._base_cls.k_and_deriv(self, X, Y=Y)
        self._locked = False
        return k, dk


class _SpinSymMixin(_IndexMixin):
    """
    A Mixin that evalutes the kernel assuming spin-symmetry
    between the alpha_ind features and beta_ind features.
    To use, make a class SpinSymClass(_SpinSymMixin, BaseClass).
    BaseClass must be the second parent.
    """

    def __init__(self, alpha_ind, beta_ind, *args, **kwargs):
        """
        Args:
            alpha_ind: list/tuple of indexes or slice for alpha spin
            beta_ind: list/tuple of indexes or slice for beta spin
            *args and **kwargs: arguments to parent kernel
        """
        self._locked = False
        self.alpha_ind = alpha_ind
        self.beta_ind = beta_ind
        self._base_cls = self.__class__.__bases__[1]  # Base Kernel
        self._base_cls.__init__(self, *args, **kwargs)

    def __call__(self, X, Y=None, eval_gradient=False, **kwargs):
        """
        Call the kernel

        Args:
            X: First matrix
            Y: Second matrix
            eval_gradient: Whether to evaluate gradient. X must be None
            **kwargs: Passed to base kernel. Primarily for using the
                get_sub_kernels option of ARBF for spline mapping.

        Returns:
            Return values of base kernel (usually kernel matrix and, if
                eval_gradient is True, its gradient wrt hyperparams)
        """
        if self._locked:
            return self._base_cls.__call__(
                self, X, Y=Y, eval_gradient=eval_gradient, **kwargs
            )
        self._locked = True
        if Y is not None:
            Y = np.vstack((Y[:, self.alpha_ind], Y[:, self.beta_ind]))
        else:
            Y = None
        X = np.vstack((X[:, self.alpha_ind], X[:, self.beta_ind]))
        NX = X.shape[0] // 2
        NY = Y.shape[0] // 2 if Y is not None else NX
        k = self._base_cls.__call__(self, X, Y=Y, eval_gradient=eval_gradient, **kwargs)
        if eval_gradient:
            k, dk = k
        k = k[:NX] + k[NX:]
        k = k[:, :NY] + k[:, NY:]
        self._locked = False
        if eval_gradient:
            dk = dk[:NX] + dk[NX:]
            dk = dk[:, :NY] + dk[:, NY:]
            return k, dk
        return k

    def diag(self, X):
        if self._locked:
            return self._base_cls.diag(self, X)
        self._locked = True
        XA = X[:, self.alpha_ind]
        XB = X[:, self.beta_ind]
        diag = self._base_cls.diag(self, XA)
        diag += self._base_cls.diag(self, XB)
        diag += np.diag(self._base_cls.__call__(self, XA, XB))
        diag += np.diag(self._base_cls.__call__(self, XB, XA))
        self._locked = False
        return diag

    def k_and_deriv(self, X, Y=None):
        if self._locked:
            return self._base_cls.k_and_deriv(self, X, Y=Y)
        self._locked = True
        Nfeat = X.shape[1]
        if Y is not None:
            Y = np.vstack((Y[:, self.alpha_ind], Y[:, self.beta_ind]))
        else:
            Y = None
        X = np.vstack((X[:, self.alpha_ind], X[:, self.beta_ind]))
        NX = X.shape[0] // 2
        NY = Y.shape[0] // 2 if Y is not None else NX
        k, dk = self._base_cls.k_and_deriv(self, X, Y=Y)
        k = k[:NX] + k[NX:]
        k = k[:, :NY] + k[:, NY:]
        dk = dk[:, :NY] + dk[:, NY:]
        dkfull = np.zeros((NX, NY, Nfeat))
        dkfull[:, :, self.alpha_ind] = dk[:NX]
        dkfull[:, :, self.beta_ind] = dk[NX:]
        self._locked = False
        return k, dkfull


class PartialRBF(DiffRBF):
    """
    Child class of sklearn RBF which only acts on the slice X[:,start:]
    (or X[:,active_dims] if and only if active_dims is supplied).
    start is ignored if active_dims is supplied.
    """

    def __init__(
        self,
        length_scale=1.0,
        length_scale_bounds=(1e-5, 1e5),
        start=0,
        active_dims=None,
    ):
        super(PartialRBF, self).__init__(length_scale, length_scale_bounds)
        self.start = start
        self.active_dims = active_dims

    def __call__(self, X, Y=None, eval_gradient=False):
        if self.active_dims is None:
            X = X[:, self.start :]
            if Y is not None:
                Y = Y[:, self.start :]
        else:
            X = X[:, self.active_dims]
            if Y is not None:
                Y = Y[:, self.active_dims]
        return super(PartialRBF, self).__call__(X, Y, eval_gradient)

    def k_and_deriv(self, X, Y=None):
        if self.active_dims is None:
            X = X[:, self.start :]
            if Y is not None:
                Y = Y[:, self.start :]
        else:
            X = X[:, self.active_dims]
            if Y is not None:
                Y = Y[:, self.active_dims]
        return super(PartialRBF, self).k_and_deriv(X, Y)


class DiffARBF(DiffRBF):
    """
    Additive RBF kernel of Duvenaud et al.
    """

    def __init__(
        self,
        order=1,
        length_scale=1.0,
        scale=None,
        length_scale_bounds=(1e-5, 1e5),
        scale_bounds=(1e-5, 1e5),
    ):
        """
        Args:
            order (int): Order of kernel
            length_scale (float or array): length scale of kernel
            scale (array): coefficients of each order, starting with
                0 and ascending.
            length_scale_bounds: bounds of length_scale
            scale_bounds: bounds of scale
        """
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds
        self.order = order
        self.scale = scale
        self.scale_bounds = scale_bounds

    @property
    def hyperparameter_scale(self):
        return Hyperparameter("scale", "numeric", self.scale_bounds, len(self.scale))

    def diag(self, X):
        nfeat = X.shape[1]
        comb_list = []
        for n in range(self.order + 1):
            comb_list.append(comb(nfeat, n))
        comb_list = np.array(comb_list)
        return np.ones(X.shape[0]) * np.sum(self.scale * comb_list)

    def __call__(self, X, Y=None, eval_gradient=False, get_sub_kernels=False):
        if self.anisotropic:
            num_scale = len(self.length_scale)
        else:
            num_scale = 1
        if Y is None:
            Y = X
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        sk = []
        if eval_gradient:
            deriv_size = 0
            if not self.hyperparameter_length_scale.fixed:
                deriv_size += num_scale
            else:
                num_scale = 0
            if not self.hyperparameter_scale.fixed:
                deriv_size += len(self.scale)
            derivs = np.zeros((X.shape[0], Y.shape[0], deriv_size))
        for i in range(self.order):
            sk.append(np.sum(k0 ** (i + 1), axis=-1))
        en = [1]
        for n in range(self.order):
            en.append(sk[n] * (-1) ** n)
            for k in range(n):
                en[-1] += (-1) ** k * en[n - k] * sk[k]
            en[-1] /= n + 1
        res = 0
        for n in range(self.order + 1):
            res += self.scale[n] * en[n]
            if eval_gradient and not self.hyperparameter_scale.fixed:
                derivs[:, :, num_scale + n] = self.scale[n] * en[n]
        kernel = res
        if get_sub_kernels:
            return kernel, en

        if eval_gradient and not self.hyperparameter_length_scale.fixed:
            inds = np.arange(X.shape[1])
            for ind in inds:
                den = [np.zeros(X.shape)]
                sktmp = []
                if self.order > 0:
                    den.append(np.ones(k0[:, :, ind].shape))
                for n in range(1, self.order):
                    sktmp.append(
                        np.sum(k0[..., :ind] ** n, axis=-1)
                        + np.sum(k0[..., ind + 1 :] ** n, axis=-1)
                    )
                    den.append(sktmp[n - 1] * (-1) ** (n + 1))
                    for k in range(n - 1):
                        den[-1] += (-1) ** k * den[n - k] * sktmp[k]
                res = 0
                if self.order > 0:
                    res += self.scale[1] * den[1]
                for n in range(2, self.order + 1):
                    res += self.scale[n] * den[n] / (n - 1)
                res *= diff[:, :, ind] ** 2 * k0[:, :, ind]
                if self.anisotropic:
                    derivs[:, :, ind] = res
                else:
                    derivs[:, :, 0] += res

        if eval_gradient:
            return kernel, derivs
        return kernel

    def k_and_deriv(self, X, Y=None):
        if Y is None:
            Y = X
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        sk = []
        for i in range(self.order):
            sk.append(np.sum(k0 ** (i + 1), axis=-1))
        en = [1]
        for n in range(self.order):
            en.append(sk[n] * (-1) ** n)
            for k in range(n):
                en[-1] += (-1) ** k * en[n - k] * sk[k]
            en[-1] /= n + 1
        res = 0
        for n in range(self.order + 1):
            res += self.scale[n] * en[n]
        kernel = res

        shape = kernel.shape + (X.shape[1],)
        dk = np.zeros(shape, dtype=kernel.dtype)
        for ind in range(X.shape[1]):
            den = [np.zeros(X.shape)]
            sktmp = []
            if self.order > 0:
                den.append(np.ones(k0[:, :, ind].shape))
            for n in range(1, self.order):
                sktmp.append(
                    np.sum(k0[..., :ind] ** n, axis=-1)
                    + np.sum(k0[..., ind + 1 :] ** n, axis=-1)
                )
                den.append(sktmp[n - 1] * (-1) ** (n + 1))
                for k in range(n - 1):
                    den[-1] += (-1) ** k * den[n - k] * sktmp[k]
            res = 0
            if self.order > 0:
                res += self.scale[1] * den[1]
            for n in range(2, self.order + 1):
                res += self.scale[n] * den[n] / (n - 1)
            res *= -1 * diff[:, :, ind] * k0[:, :, ind]
            dk[:, :, ind] = res
        dk /= self.length_scale

        return kernel, dk


class DiffAdditiveMixin(DiffKernelMixin):
    """
    Need to define is_stationary, _get_k0_dk0_train,
    and _get_k0_dk0_eval.
    """

    def __init__(
        self,
        order=1,
        length_scale=1.0,
        scale=None,
        length_scale_bounds=(1e-5, 1e5),
        scale_bounds=(1e-5, 1e5),
    ):
        """
        Args:
            order (int): Order of kernel
            length_scale (float or array): length scale of kernel
            scale (array): coefficients of each order, starting with
                0 and ascending.
            length_scale_bounds: bounds of length_scale
            scale_bounds: bounds of scale
        """
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds
        self.order = order
        self.scale = scale
        self.scale_bounds = scale_bounds

    @property
    def anisotropic(self):
        return np.iterable(self.length_scale) and len(self.length_scale) > 1

    @property
    def hyperparameter_length_scale(self):
        if self.anisotropic:
            return Hyperparameter(
                "length_scale",
                "numeric",
                self.length_scale_bounds,
                len(self.length_scale),
            )
        return Hyperparameter("length_scale", "numeric", self.length_scale_bounds)

    def is_stationary(self):
        raise NotImplementedError

    def diag(self, X):
        """
        NOTE: Quite inefficient, but this is not used much for the
        functionals anyway.
        """
        K = self(X)
        return np.diag(K)

    def get_k0_for_mapping(self, X, Y, lscale):
        raise NotImplementedError

    def _get_k0_dk0_train(self, X, Y, eval_gradient):
        raise NotImplementedError

    def _get_k0_dk0_eval(self, X, Y, eval_gradient):
        raise NotImplementedError

    @property
    def num_scale(self):
        if self.anisotropic:
            return len(self.length_scale)
        else:
            return 1

    @property
    def hyperparameter_scale(self):
        return Hyperparameter("scale", "numeric", self.scale_bounds, len(self.scale))

    def get_zero_derivs(self, X, Y):
        deriv_size = 0
        if not self.hyperparameter_length_scale.fixed:
            deriv_size += self.num_scale
        if not self.hyperparameter_scale.fixed:
            deriv_size += len(self.scale)
        return np.zeros((X.shape[0], Y.shape[0], deriv_size))

    def __call__(self, X, Y=None, eval_gradient=False, get_sub_kernels=False):
        if Y is None:
            Y = X
        k0, dk0 = self._get_k0_dk0_train(X, Y, eval_gradient)
        if eval_gradient:
            derivs = self.get_zero_derivs(X, Y)
        sk = []
        for i in range(self.order):
            sk.append(np.sum(k0 ** (i + 1), axis=-1))
        en = [1]
        for n in range(self.order):
            en.append(sk[n] * (-1) ** n)
            for k in range(n):
                en[-1] += (-1) ** k * en[n - k] * sk[k]
            en[-1] /= n + 1
        res = 0
        for n in range(self.order + 1):
            res += self.scale[n] * en[n]
            if eval_gradient and not self.hyperparameter_scale.fixed:
                derivs[:, :, self.num_scale + n] = self.scale[n] * en[n]
        kernel = res
        if get_sub_kernels:
            return kernel, en

        if eval_gradient and not self.hyperparameter_length_scale.fixed:
            inds = np.arange(X.shape[1])
            for ind in inds:
                den = [np.zeros(X.shape)]
                sktmp = []
                if self.order > 0:
                    den.append(np.ones(k0[:, :, ind].shape))
                for n in range(1, self.order):
                    sktmp.append(
                        np.sum(k0[..., :ind] ** n, axis=-1)
                        + np.sum(k0[..., ind + 1 :] ** n, axis=-1)
                    )
                    den.append(sktmp[n - 1] * (-1) ** (n + 1))
                    for k in range(n - 1):
                        den[-1] += (-1) ** k * den[n - k] * sktmp[k]
                res = 0
                if self.order > 0:
                    res += self.scale[1] * den[1]
                for n in range(2, self.order + 1):
                    res += self.scale[n] * den[n] / (n - 1)
                res *= dk0[:, :, ind]
                if self.anisotropic:
                    derivs[:, :, ind] = res
                else:
                    derivs[:, :, 0] += res

        if eval_gradient:
            return kernel, derivs
        return kernel

    def k_and_deriv(self, X, Y=None):
        if Y is None:
            Y = X
        k0, dk0 = self._get_k0_dk0_eval(X, Y, True)
        sk = []
        for i in range(self.order):
            sk.append(np.sum(k0 ** (i + 1), axis=-1))
        en = [1]
        for n in range(self.order):
            en.append(sk[n] * (-1) ** n)
            for k in range(n):
                en[-1] += (-1) ** k * en[n - k] * sk[k]
            en[-1] /= n + 1
        res = 0
        for n in range(self.order + 1):
            res += self.scale[n] * en[n]
        kernel = res

        shape = kernel.shape + (X.shape[1],)
        dk = np.zeros(shape, dtype=kernel.dtype)
        for ind in range(X.shape[1]):
            den = [np.zeros(X.shape)]
            sktmp = []
            if self.order > 0:
                den.append(np.ones(k0[:, :, ind].shape))
            for n in range(1, self.order):
                sktmp.append(
                    np.sum(k0[..., :ind] ** n, axis=-1)
                    + np.sum(k0[..., ind + 1 :] ** n, axis=-1)
                )
                den.append(sktmp[n - 1] * (-1) ** (n + 1))
                for k in range(n - 1):
                    den[-1] += (-1) ** k * den[n - k] * sktmp[k]
            res = 0
            if self.order > 0:
                res += self.scale[1] * den[1]
            for n in range(2, self.order + 1):
                res += self.scale[n] * den[n] / (n - 1)
            res *= dk0[:, :, ind]
            dk[:, :, ind] = res

        return kernel, dk


class DiffARBFV2(DiffAdditiveMixin, Kernel):
    """
    Additive RBF kernel of Duvenaud et al.
    """

    def is_stationary(self):
        return True

    def get_k0_for_mapping(self, X, Y, lscale):
        diff = X[:, None] - Y[None, :] / lscale
        return np.exp(-0.5 * diff * diff)

    def _get_k0_dk0_train(self, X, Y, eval_gradient):
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        if eval_gradient:
            dk0 = diff**2 * k0
        else:
            dk0 = None
        return k0, dk0

    def _get_k0_dk0_eval(self, X, Y, eval_gradient):
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        if eval_gradient:
            dk0 = -diff * k0 / self.length_scale
        else:
            dk0 = None
        return k0, dk0


class DiffAddLLRBF(DiffAdditiveMixin, Kernel):
    """
    Additive RBF kernel of Duvenaud et al.
    """

    def __init__(
        self,
        order=1,
        alpha=1.0,
        length_scale=1.0,
        scale=None,
        length_scale_bounds=(1e-5, 1e5),
        scale_bounds=(1e-5, 1e5),
    ):
        """
        Args:
            order (int): Order of kernel
            alpha (float): Rational quadratic order
            length_scale (float or array): length scale of kernel
            scale (array): coefficients of each order, starting with
                0 and ascending.
            length_scale_bounds: bounds of length_scale
            scale_bounds: bounds of scale
        """
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds
        self.order = order
        self.alpha = alpha
        self.scale = scale
        self.scale_bounds = scale_bounds

    def is_stationary(self):
        return False

    def get_k0_for_mapping(self, X, Y, lscale):
        diff = (X[:, np.newaxis] - Y[np.newaxis, :]) / lscale
        dot = (X[:, np.newaxis] * Y[np.newaxis, :]) / (self.alpha * lscale**2)
        return (1 + dot) * np.exp(-0.5 * diff**2)

    def _get_k0_dk0_train(self, X, Y, eval_gradient):
        invssq = 1.0 / (self.alpha * self.length_scale**2)
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        dot = 1 + invssq * X[:, np.newaxis, :] * Y[np.newaxis, :, :]
        if eval_gradient:
            dk0 = diff**2 * k0
            ddot = -2 * invssq * X[:, np.newaxis, :] * Y[np.newaxis, :, :]
            dk0 = dk0 * dot + ddot * k0
        else:
            dk0 = None
        k0 = dot * k0
        return k0, dk0

    def _get_k0_dk0_eval(self, X, Y, eval_gradient):
        invssq = 1.0 / (self.alpha * self.length_scale**2)
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        dot = 1 + invssq * X[:, np.newaxis, :] * Y[np.newaxis, :, :]
        if eval_gradient:
            dk0 = -1.0 * diff * k0 / self.length_scale
            ddot = invssq * Y[np.newaxis, :, :]
            dk0 = dk0 * dot + ddot * k0
        else:
            dk0 = None
        k0 = dot * k0
        return k0, dk0


class DiffAddRQ(DiffAdditiveMixin, Kernel):
    """
    Additive rational quadratic kernel
    """

    def __init__(
        self,
        order=1,
        alpha=1.0,
        length_scale=1.0,
        scale=None,
        length_scale_bounds=(1e-5, 1e5),
        scale_bounds=(1e-5, 1e5),
    ):
        """
        Args:
            order (int): Order of kernel
            alpha (float): Rational quadratic order
            length_scale (float or array): length scale of kernel
            scale (array): coefficients of each order, starting with
                0 and ascending.
            length_scale_bounds: bounds of length_scale
            scale_bounds: bounds of scale
        """
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds
        self.order = order
        self.alpha = alpha
        self.scale = scale
        self.scale_bounds = scale_bounds

    def is_stationary(self):
        return False

    def get_k0_for_mapping(self, X, Y, lscale):
        diff = X[:, np.newaxis] - Y[np.newaxis, :]
        inv_scale = 1.0 / (2 * self.alpha * lscale**2)
        return (1 + diff * diff * inv_scale) ** (-self.alpha)

    def _get_k0_dk0_train(self, X, Y, eval_gradient):
        diff = X[:, np.newaxis, :] - Y[np.newaxis, :, :]
        inv_scale = 1.0 / (2 * self.alpha * self.length_scale**2)
        alpha = self.alpha
        k0 = (1 + diff * diff * inv_scale) ** (-alpha)
        if eval_gradient:
            dk0 = (
                2
                * diff
                * diff
                * (1 + diff * diff * inv_scale) ** (-1 - alpha)
                * inv_scale
                * alpha
            )
        else:
            dk0 = None
        return k0, dk0

    def _get_k0_dk0_eval(self, X, Y, eval_gradient):
        diff = X[:, np.newaxis, :] - Y[np.newaxis, :, :]
        inv_scale = 1.0 / (2 * self.alpha * self.length_scale**2)
        alpha = self.alpha
        k0 = (1 + diff * diff * inv_scale) ** (-alpha)
        if eval_gradient:
            dk0 = (
                -2
                * diff
                * (1 + diff * diff * inv_scale) ** (-1 - alpha)
                * inv_scale
                * alpha
            )
        else:
            dk0 = None
        return k0, dk0


def qarbf_args(arbf_base):
    ndim = len(arbf_base.length_scale)
    length_scale = arbf_base.length_scale
    scale = [arbf_base.scale[0]]
    scale += [arbf_base.scale[1]] * (ndim)
    scale += [arbf_base.scale[2]] * (ndim * (ndim - 1) // 2)
    return ndim, np.array(length_scale), scale


def arbf_args(arbf_base):
    ndim = len(arbf_base.length_scale)
    length_scale = arbf_base.length_scale
    order = arbf_base.order
    scale = [arbf_base.scale[0]]
    if order > 0:
        scale += [arbf_base.scale[1]] * (ndim)
    if order > 1:
        scale += [arbf_base.scale[2]] * (ndim * (ndim - 1) // 2)
    if order > 2:
        scale += [arbf_base.scale[3]] * (ndim * (ndim - 1) * (ndim - 2) // 6)
    if order > 3:
        raise ValueError("Order too high for mapping")
    return ndim, np.array(length_scale), scale, order


class QARBF(StationaryKernelMixin, Kernel):
    """
    ARBF, except order is restricted to 2 and
    the algorithm is more efficient.
    """

    def __init__(self, ndim, length_scale, scale, scale_bounds=(1e-5, 1e5)):
        super(QARBF, self).__init__()
        self.ndim = ndim
        self.scale = scale
        self.scale_bounds = scale_bounds
        self.length_scale = length_scale

    @property
    def hyperparameter_scale(self):
        return Hyperparameter("scale", "numeric", self.scale_bounds, len(self.scale))

    def diag(self, X):
        return np.diag(self.__call__(X))

    def __call__(self, X, Y=None, eval_gradient=False):
        if Y is None:
            Y = X
        diff = (X[:, np.newaxis, :] - Y[np.newaxis, :, :]) / self.length_scale
        k0 = np.exp(-0.5 * diff**2)
        sk = np.zeros((X.shape[0], Y.shape[0], len(self.scale)))
        sk[:, :, 0] = self.scale[0]
        t = 1
        for i in range(self.ndim):
            sk[:, :, t] = self.scale[t] * k0[:, :, i]
            t += 1
        for i in range(self.ndim - 1):
            for j in range(i + 1, self.ndim):
                sk[:, :, t] = self.scale[t] * k0[:, :, i] * k0[:, :, j]
                t += 1
        k = np.sum(sk, axis=-1)
        print(self.scale)
        if eval_gradient:
            return k, sk
        return k

    def get_sub_kernel(self, inds, scale_ind, X, Y):
        if inds is None:
            return self.scale[0] * np.ones((X.shape[0], Y.shape[0]))
        if isinstance(inds, int):
            diff = (X[:, np.newaxis, inds] - Y[np.newaxis, :]) / self.length_scale[inds]
            k0 = np.exp(-0.5 * diff**2)
            return self.scale[scale_ind] * k0
        else:
            diff = (
                X[:, np.newaxis, inds[0]] - Y[np.newaxis, :, 0]
            ) / self.length_scale[inds[0]]
            k0 = np.exp(-0.5 * diff**2)
            diff = (
                X[:, np.newaxis, inds[1]] - Y[np.newaxis, :, 1]
            ) / self.length_scale[inds[1]]
            k0 *= np.exp(-0.5 * diff**2)
            return self.scale[scale_ind] * k0

    def get_funcs_for_spline_conversion(self):
        funcs = [lambda x, y: self.get_sub_kernel(None, 0, x, y)]
        t = 1
        for i in range(self.ndim):
            funcs.append(lambda x, y: self.get_sub_kernel(i, t, x, y))
            t += 1
        for i in range(self.ndim - 1):
            for j in range(i + 1, self.ndim):
                funcs.append(lambda x, y: self.get_sub_kernel((i, j), t, x, y))
                t += 1
        return funcs


class PartialARBF(DiffARBF):
    """
    ARBF where subset of X is selected.
    """

    def __init__(
        self,
        order=1,
        length_scale=1.0,
        length_scale_bounds=(1e-5, 1e5),
        scale=1.0,
        scale_bounds=(1e-5, 1e5),
        start=1,
        active_dims=None,
    ):
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds
        self.order = order
        self.scale = scale  # if (scale is None) else [1.0] * (order + 1)
        self.scale_bounds = scale_bounds
        self.start = start
        self.active_dims = active_dims

    def __call__(self, X, Y=None, eval_gradient=False, get_sub_kernels=False):
        # hasattr check for back-compatibility
        if not np.iterable(self.scale):
            self.scale = [self.scale] * (self.order + 1)
        if (not hasattr(self, "active_dims")) or (self.active_dims is None):
            X = X[:, self.start :]
            if Y is not None:
                Y = Y[:, self.start :]
        else:
            X = X[:, self.active_dims]
            if Y is not None:
                Y = Y[:, self.active_dims]
        return super(PartialARBF, self).__call__(X, Y, eval_gradient, get_sub_kernels)


class SingleRBF(RBF):
    """
    RBF kernel with single index of X selected.
    """

    def __init__(self, length_scale=1.0, length_scale_bounds=(1e-5, 1e5), index=1):
        super(SingleRBF, self).__init__(length_scale, length_scale_bounds)
        self.index = index

    def __call__(self, X, Y=None, eval_gradient=False):

        X = X[:, self.index : self.index + 1]
        if Y is not None:
            Y = Y[:, self.index : self.index + 1]
        return super(SingleRBF, self).__call__(X, Y, eval_gradient)


class SingleDot(DotProduct):
    """
    DotProduct kernel with single index of X selected.
    """

    def __init__(self, sigma_0=1.0, sigma_0_bounds=(1e-05, 100000.0), index=0):
        super(SingleDot, self).__init__(sigma_0, sigma_0_bounds)
        self.index = index

    def __call__(self, X, Y=None, eval_gradient=False):
        X = X[:, self.index : self.index + 1]
        if Y is not None:
            Y = Y[:, self.index : self.index + 1]
        return super(SingleDot, self).__call__(X, Y, eval_gradient)


class DensityNoise(StationaryKernelMixin, GenericKernelMixin, Kernel):
    def __init__(self, index=0):
        self.index = index

    def __call__(self, X, Y=None, eval_gradient=False):
        if Y is not None and eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        if Y is None:
            K = np.diag(self.diag(X))
            if eval_gradient:
                grad = np.empty((_num_samples(X), _num_samples(X), 0))
                return K, grad
            else:
                return K
        else:
            return np.zeros((_num_samples(X), _num_samples(Y)))

    def diag(self, X):
        return 1 / X[:, self.index] ** 2


class ExponentialDensityNoise(StationaryKernelMixin, GenericKernelMixin, Kernel):
    def __init__(self, exponent=1.0, exponent_bounds=(0.1, 10)):
        self.exponent = exponent
        self.exponent_bounds = exponent_bounds

    @property
    def hyperparameter_exponent(self):
        return Hyperparameter("exponent", "numeric", self.exponent_bounds)

    def __call__(self, X, Y=None, eval_gradient=False):
        if Y is not None and eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        if Y is None:
            K = np.diag(self.diag(X))
            if eval_gradient and not self.hyperparameter_exponent.fixed:
                rho = X[:, 0]
                grad = np.empty((_num_samples(X), _num_samples(X), 1))
                grad[:, :, 0] = np.diag(
                    -self.exponent * np.log(rho) / rho**self.exponent
                )
                return K, grad
            elif eval_gradient:
                grad = np.zeros((X.shape[0], X.shape[0], 0))
                return K, grad
            else:
                return K
        else:
            return np.zeros((_num_samples(X), _num_samples(Y)))

    def diag(self, X):
        return 1 / X[:, 0] ** self.exponent

    def __repr__(self):
        return "{0}(exponent={1:.3g})".format(self.__class__.__name__, self.exponent)


class FittedDensityNoise(StationaryKernelMixin, GenericKernelMixin, Kernel):
    """
    Kernel to model the noise of the exchange enhancement factor based
    on the density. 1 / (1 + decay_rate * rho)
    """

    def __init__(self, decay_rate=4.0, decay_rate_bounds=(1e-5, 1e5)):
        self.decay_rate = decay_rate
        self.decay_rate_bounds = decay_rate_bounds

    @property
    def hyperparameter_decay_rate(self):
        return Hyperparameter("decay_rate", "numeric", self.decay_rate_bounds)

    def __call__(self, X, Y=None, eval_gradient=False):
        if Y is not None and eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        if Y is None:
            K = np.diag(self.diag(X))
            if eval_gradient and not self.hyperparameter_decay_rate.fixed:
                rho = X[:, 0]
                grad = np.empty((_num_samples(X), _num_samples(X), 1))
                grad[:, :, 0] = np.diag(-rho / (1 + self.decay_rate * rho) ** 2)
                return K, grad
            elif eval_gradient:
                grad = np.zeros((X.shape[0], X.shape[0], 0))
                return K, grad
            else:
                return K
        else:
            return np.zeros((_num_samples(X), _num_samples(Y)))

    def diag(self, X):
        rho = X[:, 0]
        return 1 / (1 + self.decay_rate * rho)

    def __repr__(self):
        return "{0}(decay_rate={1:.3g})".format(
            self.__class__.__name__, self.decay_rate
        )


class ADKernel(Kernel):
    def __init__(self, k, active_dims):
        self.k = k
        self.active_dims = active_dims

    def get_params(self, deep=True):
        params = dict(k=self.k, active_dims=self.active_dims)
        if deep:
            deep_items = self.k.get_params().items()
            params.update(("k__" + k, val) for k, val in deep_items)

        return params

    @property
    def hyperparameters(self):
        """Returns a list of all hyperparameter."""
        return [
            Hyperparameter(
                "k__" + hyperparameter.name,
                hyperparameter.value_type,
                hyperparameter.bounds,
                hyperparameter.n_elements,
            )
            for hyperparameter in self.k.hyperparameters
        ]

    @property
    def theta(self):
        """Returns the (flattened, log-transformed) non-fixed hyperparameters.
        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.
        Returns
        -------
        theta : ndarray of shape (n_dims,)
            The non-fixed, log-transformed hyperparameters of the kernel
        """
        return self.k.theta

    @theta.setter
    def theta(self, theta):
        """Sets the (flattened, log-transformed) non-fixed hyperparameters.
        Parameters
        ----------
        theta : ndarray of shape (n_dims,)
            The non-fixed, log-transformed hyperparameters of the kernel
        """
        self.k.theta = theta

    @property
    def bounds(self):
        """Returns the log-transformed bounds on the theta.
        Returns
        -------
        bounds : ndarray of shape (n_dims, 2)
            The log-transformed bounds on the kernel's hyperparameters theta
        """
        return self.k.bounds

    def __eq__(self, b):
        return self.k == b.k and self.active_dims == b.active_dims

    def is_stationary(self):
        """Returns whether the kernel is stationary."""
        return self.k.is_stationary()

    @property
    def requires_vector_input(self):
        """Returns whether the kernel is stationary."""
        return self.k.requires_vector_input

    def __call__(self, X, Y=None, eval_gradient=False):
        X = X[:, self.active_dims]
        if Y is not None:
            Y = Y[:, self.active_dims]
        return self.k.__call__(X, Y, eval_gradient)

    def diag(self, X):
        return self.k.diag(X)

    def __repr__(self):
        return self.k.__repr__()


class SpinSymKernel(ADKernel):
    def __init__(self, k, up_active_dims, down_active_dims):
        self.k = k
        self.up_active_dims = up_active_dims
        self.down_active_dims = down_active_dims

    def get_params(self, deep=True):
        params = dict(
            k=self.k,
            up_active_dims=self.up_active_dims,
            down_active_dims=self.down_active_dims,
        )
        if deep:
            deep_items = self.k.get_params().items()
            params.update(("k__" + k, val) for k, val in deep_items)

        return params

    def __call__(self, X, Y=None, eval_gradient=False):
        Xup = X[:, self.up_active_dims]
        if Y is not None:
            Yup = Y[:, self.up_active_dims]
        else:
            Yup = None
        kup = self.k.__call__(Xup, Yup, eval_gradient)
        Xdown = X[:, self.down_active_dims]
        if Y is not None:
            Ydown = Y[:, self.down_active_dims]
        else:
            Ydown = None
        kdown = self.k.__call__(Xdown, Ydown, eval_gradient)
        if eval_gradient:
            return kup[0] + kdown[0], kup[1] + kdown[1]
        else:
            return kup + kdown


class SubsetRBF(_SubsetMixin, DiffRBF):
    pass


class SubsetARBF(_SubsetMixin, DiffARBF):
    pass


class SubsetAddLLRBF(_SubsetMixin, DiffAddLLRBF):
    pass


class SubsetAddRQ(_SubsetMixin, DiffAddRQ):
    pass


class SubsetPoly(_SubsetMixin, DiffPolyKernel):
    pass


class SpinSymRBF(_SpinSymMixin, DiffRBF):
    pass


class SpinSymARBF(_SpinSymMixin, DiffARBF):
    pass


class SpinSymPoly(_SpinSymMixin, DiffPolyKernel):
    pass



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

# TODO not using wheel yet, but plan to do so eventually
from wheel.bdist_wheel import bdist_wheel


def get_version():
    topdir = os.path.abspath(os.path.join(__file__, ".."))
    with open(os.path.join(topdir, "ciderpress", "__init__.py"), "r") as f:
        for line in f.readlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise ValueError("Version string not found")


VERSION = get_version()


def get_platform():
    from sysconfig import get_platform

    platform = get_platform()
    # TODO might want to add darwin OSX support like PySCF
    # but only after officially adding OSX support
    return platform


class CMakeBuildPy(build_py):
    def run(self):
        self.plat_name = get_platform()
        self.build_base = "build"
        self.build_lib = os.path.join(self.build_base, "lib")
        self.build_temp = os.path.join(self.build_base, f"temp.{self.plat_name}")

        self.announce("Configuring extensions", level=3)
        src_dir = os.path.abspath(os.path.join(__file__, "..", "ciderpress", "lib"))
        cmd = [
            "cmake",
            f"-S{src_dir}",
            f"-B{self.build_temp}",
            "-DCMAKE_PREFIX_PATH={}".format(sys.base_prefix),
            "-DBLA_VENDOR=Intel10_64lp_seq",
            "-DCMAKE_BUILD_TYPE=Release",
        ]
        configure_args = os.getenv("CMAKE_CONFIGURE_ARGS")
        if configure_args:
            cmd.extend(configure_args.split(" "))
        self.spawn(cmd)

        self.announce("Building binaries", level=3)
        cmd = ["cmake", "--build", self.build_temp, "-j2"]
        build_args = os.getenv("CMAKE_BUILD_ARGS")
        if build_args:
            cmd.extend(build_args.split(" "))
        if self.dry_run:
            self.announce(" ".join(cmd))
        else:
            self.spawn(cmd)
        self.editable_mode = False
        super().run()


# NOTE note trying to support wheal yet, but including
# these code block from PySCF setup.py for future.
initialize_options_1 = bdist_wheel.initialize_options


def initialize_with_default_plat_name(self):
    initialize_options_1(self)
    self.plat_name = get_platform()
    self.plat_name_supplied = True


bdist_wheel.initialize_options = initialize_with_default_plat_name

# from PySCF setup.py
try:
    from setuptools.command.bdist_wheel import bdist_wheel

    initialize_options_2 = bdist_wheel.initialize_options

    def initialize_with_default_plat_name(self):
        initialize_options_2(self)
        self.plat_name = get_platform()
        self.plat_name_supplied = True

    bdist_wheel.initialize_options = initialize_with_default_plat_name
except ImportError:
    pass

setup(
    version=VERSION,
    include_package_data=True,
    packages=find_packages(exclude=["*test*", "*examples*"]),
    cmdclass={"build_py": CMakeBuildPy},
)



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np

from ciderpress.dft.settings import LDA_FACTOR
from ciderpress.lib import load_library

xc_helper = load_library("libxc_utils")


def nsp_rho_basline(X0T, e=None, dedx=None):
    nspin, nfeat, nsamp = X0T.shape
    if e is None:
        e = np.zeros(nsamp)
    if dedx is None:
        dedx = np.zeros_like(X0T)
    nspin = X0T.shape[0]
    e[:] += X0T[:, 0].mean(0)
    dedx[0, :] += 1.0 / nspin


def _lda_x_helper(X0T, e, dedx):
    rho = X0T[0]
    e[:] += LDA_FACTOR * rho ** (4.0 / 3)
    dedx[0] += 4.0 / 3 * LDA_FACTOR * rho ** (1.0 / 3)


def _vi_x_damp_helper(X0T, e, dedx):
    rho = X0T[0]
    damp = 2.0 / (1.0 + 0.5 * X0T[3]) ** 2
    ddamp = -1 * damp / (1.0 + 0.5 * X0T[3])
    e[:] += LDA_FACTOR * damp * rho ** (4.0 / 3)
    dedx[0] += 4.0 / 3 * LDA_FACTOR * damp * rho ** (1.0 / 3)
    dedx[3] += LDA_FACTOR * ddamp * rho ** (4.0 / 3)


def _pbe_x_helper(X0T, e, dedx):
    rho = X0T[0]
    s2 = X0T[1]
    kappa = 0.804
    mu = 0.2195149727645171
    mk = mu / kappa
    fac = 1.0 / (1 + mk * s2)
    fx = 1 + kappa - kappa * fac
    dfx = mu * fac * fac
    e[:] += LDA_FACTOR * rho ** (4.0 / 3) * fx
    dedx[0] += 4.0 / 3 * LDA_FACTOR * rho ** (1.0 / 3) * fx
    dedx[1] += LDA_FACTOR * rho ** (4.0 / 3) * dfx


def _chachiyo_x_helper(X0T, e, dedx):
    rho = X0T[0]
    s2 = X0T[1]
    c = 4 * np.pi / 9
    x = c * np.sqrt(s2)
    dx = c / (2 * np.sqrt(s2))
    chfx = (3 * x**2 + np.pi**2 * np.log(1 + x)) / (
        (np.pi**2 + 3 * x) * np.log(1 + x)
    )
    dchfx = (
        -3 * x**2 * (np.pi**2 + 3 * x)
        + 3 * x * (1 + x) * (2 * np.pi**2 + 3 * x) * np.log(1 + x)
        - 3 * np.pi**2 * (1 + x) * np.log(1 + x) ** 2
    ) / ((1 + x) * (np.pi**2 + 3 * x) ** 2 * np.log(1 + x) ** 2)
    dchfx *= dx
    chfx[s2 < 1e-8] = 1 + 8 * s2[s2 < 1e-8] / 27
    dchfx[s2 < 1e-8] = 8.0 / 27
    e[:] += LDA_FACTOR * rho ** (4.0 / 3) * chfx
    dedx[0] += 4.0 / 3 * LDA_FACTOR * rho ** (1.0 / 3) * chfx
    dedx[1] += LDA_FACTOR * rho ** (4.0 / 3) * dchfx


def _sl_x_helper(X0T, _helper):
    nspin, nfeat, nsamp = X0T.shape
    e = np.zeros(nsamp)
    dedx = np.zeros_like(X0T)
    nspin = X0T.shape[0]
    for s in range(nspin):
        _helper(X0T[s], e, dedx[s])
    e[:] /= nspin
    dedx[:] /= nspin
    return e, dedx


def zero_xc(X0T):
    nspin, nfeat, nsamp = X0T.shape
    e = np.zeros(nsamp)
    dedx = np.zeros_like(X0T)
    return e, dedx


def one_xc(X0T):
    nspin, nfeat, nsamp = X0T.shape
    e = np.ones(nsamp)
    dedx = np.zeros_like(X0T)
    return e, dedx


def lda_x(X0T):
    return _sl_x_helper(X0T, _lda_x_helper)


def nlda_x_damp(X0T):
    return _sl_x_helper(X0T, _vi_x_damp_helper)


def gga_x_chachiyo(X0T):
    return _sl_x_helper(X0T, _chachiyo_x_helper)


def gga_x_pbe(X0T):
    return _sl_x_helper(X0T, _pbe_x_helper)


def get_libxc_lda_baseline(xcid, rho):
    if isinstance(xcid, str):
        xcid = LDA_CODES[xcid]
    nspin, size = rho.shape
    rho = np.asfortranarray(rho)
    exc = np.zeros(size)
    vrho = np.zeros_like(rho, order="F")
    xc_helper.get_lda_baseline(
        ctypes.c_int(xcid),
        ctypes.c_int(nspin),
        ctypes.c_int(size),
        rho.ctypes.data_as(ctypes.c_void_p),
        exc.ctypes.data_as(ctypes.c_void_p),
        vrho.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(1e-12),
    )
    return exc, vrho


def get_libxc_gga_baseline(xcid, rho, sigma):
    if isinstance(xcid, str):
        xcid = GGA_CODES[xcid]
    nspin, size = rho.shape
    assert sigma.shape == (2 * nspin - 1, size)
    rho = np.asfortranarray(rho)
    sigma = np.asfortranarray(sigma)
    exc = np.zeros(size)
    vrho = np.zeros_like(rho, order="F")
    vsigma = np.zeros_like(sigma, order="F")
    xc_helper.get_gga_baseline(
        ctypes.c_int(xcid),
        ctypes.c_int(nspin),
        ctypes.c_int(size),
        rho.ctypes.data_as(ctypes.c_void_p),
        sigma.ctypes.data_as(ctypes.c_void_p),
        exc.ctypes.data_as(ctypes.c_void_p),
        vrho.ctypes.data_as(ctypes.c_void_p),
        vsigma.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(1e-12),
    )
    return exc, vrho, vsigma


def get_libxc_mgga_baseline(xcid, rho, sigma, tau):
    if isinstance(xcid, str):
        xcid = MGGA_CODES[xcid]
    nspin, size = rho.shape
    assert sigma.shape == (2 * nspin - 1, size)
    rho = np.asfortranarray(rho)
    # rho = np.ascontiguousarray(rho)
    sigma = np.asfortranarray(sigma)
    # sigma = np.ascontiguousarray(sigma)
    exc = np.zeros(size)
    vrho = np.zeros_like(rho, order="F")
    vsigma = np.zeros_like(sigma, order="F")
    vtau = np.zeros_like(tau, order="F")
    xc_helper.get_mgga_baseline(
        ctypes.c_int(xcid),
        ctypes.c_int(nspin),
        ctypes.c_int(size),
        rho.ctypes.data_as(ctypes.c_void_p),
        sigma.ctypes.data_as(ctypes.c_void_p),
        tau.ctypes.data_as(ctypes.c_void_p),
        exc.ctypes.data_as(ctypes.c_void_p),
        vrho.ctypes.data_as(ctypes.c_void_p),
        vsigma.ctypes.data_as(ctypes.c_void_p),
        vtau.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(1e-12),
    )
    return exc, vrho, vsigma, vtau


def get_libxc_baseline_ss(xcid, rho_tuple):
    nspin = rho_tuple[0].shape[0]
    assert xcid in SS_GGA_CODES
    if nspin == 2:
        _rho_tuple = tuple([r.copy(order="F") for r in rho_tuple])
        for r in _rho_tuple:
            r[1:] = 0.0
        exc_a, vrho_a, vsigma_a = get_libxc_gga_baseline(
            SS_GGA_CODES[xcid], _rho_tuple[0], _rho_tuple[1]
        )
        for r, ref in zip(_rho_tuple, rho_tuple):
            r[:1] = ref[-1]
        exc_b, vrho_b, vsigma_b = get_libxc_gga_baseline(
            SS_GGA_CODES[xcid], _rho_tuple[0], _rho_tuple[1]
        )
        exc = exc_a * rho_tuple[0][0] + exc_b * rho_tuple[0][1]
        vrho_a[1] = vrho_b[0]
        vsigma_a[1] = 0
        vsigma_a[2] = vsigma_b[0]
        vrho = vrho_a
        vsigma = vsigma_a
    else:
        assert nspin == 1
        ngrid = rho_tuple[0].shape[1]
        _rho_tuple = (np.zeros((2, ngrid), order="F"), np.zeros((3, ngrid), order="F"))
        _rho_tuple[0][0] = 0.5 * rho_tuple[0][0]
        _rho_tuple[1][0] = 0.25 * rho_tuple[1][0]
        exc, vrho, vsigma = get_libxc_gga_baseline(
            SS_GGA_CODES[xcid], _rho_tuple[0], _rho_tuple[1]
        )
        exc *= _rho_tuple[0][0] * 2
        vrho = vrho[:1]
        vsigma = 0.5 * vsigma[:1]
    return exc, vrho, vsigma


def get_libxc_baseline_os(xcid, rho_tuple):
    assert xcid in OS_GGA_CODES
    exc, vrho, vsigma = get_libxc_gga_baseline(
        OS_GGA_CODES[xcid], rho_tuple[0], rho_tuple[1]
    )
    exc *= rho_tuple[0].sum(0)
    ss_xcid = "SS_" + xcid[3:]
    exc_ss, vrho_ss, vsigma_ss = get_libxc_baseline_ss(ss_xcid, rho_tuple)
    exc[:] -= exc_ss
    vrho[:] -= vrho_ss
    vsigma[:] -= vsigma_ss
    return exc, vrho, vsigma


def get_libxc_baseline(xcid, rho_tuple):
    for r in rho_tuple:
        assert r.ndim == 2
    if xcid in LDA_CODES:
        res = get_libxc_lda_baseline(xcid, rho_tuple[0])
    elif xcid in GGA_CODES:
        res = get_libxc_gga_baseline(xcid, rho_tuple[0], rho_tuple[1])
    elif xcid in MGGA_CODES:
        res = get_libxc_mgga_baseline(xcid, rho_tuple[0], rho_tuple[1], rho_tuple[2])
    elif xcid in SS_GGA_CODES:
        return get_libxc_baseline_ss(xcid, rho_tuple)
    elif xcid in OS_GGA_CODES:
        return get_libxc_baseline_os(xcid, rho_tuple)
    else:
        raise ValueError("Unsupported xcid {}".format(xcid))
    res[0][:] *= rho_tuple[0].sum(0)
    return res


def get_sigma(X0T):
    const = 4 * (3 * np.pi**2) ** (2.0 / 3)
    rho = X0T[:, 0]
    p = X0T[:, 1]
    sigma_s = const * p * rho ** (8.0 / 3)
    nspin = X0T.shape[0]
    if nspin == 1:
        return rho, sigma_s
    else:
        assert nspin == 2
        sigma = np.zeros((3, X0T.shape[-1]), order="F")
        zeta = (rho[0] - rho[1]) / (rho[0] + rho[1])
        zfac = 0.5 * (1 - zeta * zeta) / (1 + zeta * zeta)
        sigma[0] = sigma_s[0]
        sigma[2] = sigma_s[1]
        sigma[1] = (sigma_s[0] + sigma_s[1]) * zfac
        return 0.5 * rho, 0.25 * sigma


def get_dsigma(X0T, vX0T, vrho, vsigma):
    const = 4 * (3 * np.pi**2) ** (2.0 / 3)
    rho = X0T[:, 0]
    p = X0T[:, 1]
    sigma_s = const * p * rho ** (8.0 / 3)
    nspin = X0T.shape[0]
    if nspin == 1:
        vsigma_s = vsigma
    else:
        assert nspin == 2
        vrho *= 0.5
        vsigma *= 0.25
        zeta = (rho[0] - rho[1]) / (rho[0] + rho[1])
        zfac = 0.5 * (1 - zeta * zeta) / (1 + zeta * zeta)
        dzfac = -2.0 * zeta / (1 + zeta * zeta) ** 2
        vsigma_s = vsigma[::2] + vsigma[1] * zfac
        vzfac = vsigma[1] * (sigma_s[0] + sigma_s[1]) * dzfac
        vX0T[0, 0] += vzfac * 2 * rho[1] / (rho[0] + rho[1]) ** 2
        vX0T[1, 0] -= vzfac * 2 * rho[0] / (rho[0] + rho[1]) ** 2
    vX0T[:, 0] += vrho + vsigma_s * const * p * (8.0 / 3) * rho ** (5.0 / 3)
    vX0T[:, 1] += vsigma_s * const * rho ** (8.0 / 3)


def get_gga_c(xcid, X0T):
    dedx = np.zeros_like(X0T)
    rho, sigma = get_sigma(X0T)
    X0T.shape[0]
    e, vrho, vsigma = get_libxc_gga_baseline(xcid, rho, sigma)
    get_dsigma(X0T, dedx, vrho, vsigma)
    return e * rho.sum(0), dedx


def gga_c_pbe(X0T):
    return get_gga_c(130, X0T)


def mgga_c_r2scan(X0T):
    raise NotImplementedError


BASELINE_CODES = {
    "RHO": nsp_rho_basline,
    "ZERO": zero_xc,
    "ONE": one_xc,
    "LDA_X": lda_x,
    "NLDA_X_DAMP": nlda_x_damp,
    "GGA_X_PBE": gga_x_pbe,
    "GGA_X_CHACHIYO": gga_x_chachiyo,
    "GGA_C_PBE": gga_c_pbe,
}

LDA_CODES = {
    "LDA_X": 1,
    "LDA_C_PW_MOD": 13,
}

GGA_CODES = {
    "GGA_X_PBE": 101,
    "GGA_C_PBE": 130,
    "GGA_X_PBE_SOL": 116,
    "GGA_C_PBE_SOL": 133,
}

MGGA_CODES = {
    "MGGA_X_R2SCAN": 497,
    "MGGA_C_R2SCAN": 498,
}

OS_GGA_CODES = {
    "OS_GGA_C_PBE": 130,
}

OS_MGGA_CODES = {
    "OS_MGGA_C_R2SCAN": 498,
}

SS_GGA_CODES = {
    "SS_GGA_C_PBE": 130,
}

SS_MGGA_CODES = {
    "SS_MGGA_C_R2SCAN": 498,
}
